`define arb_state_wait4request	4'b0000
`define arb_state_wait4cpu	4'b0001
`define arb_state_read1		4'b0010
`define arb_state_read2		4'b0011
`define arb_state_read3		4'b0100
`define arb_state_write1	4'b0101
`define arb_state_write2	4'b0110
`define arb_state_write3	4'b0111
`define arb_state_finish	4'b1000

`define LO		1'b0
`define HI		1'b1

`define select_cpu	1'b0
`define select_capim	1'b1

// Arbiter between memory accesses from CPU and Capim-Memory-Interface
module memory_arbiter(
		      clk,
		      reset,
		      
		      // Signals between arbiter and mp_mem_interface
		      MAd_arb,
		      MDat_in_arb,
		      MDat_out_arb,
		      req_arb,
		      reqp_arb,
		      BRdy_arb,
		      wr_arb,
		      burst_arb,
		      
		      // Signals between arbiter and CPU
		      MAd_cpu,
		      MDat_in_cpu,
		      MDat_out_cpu,
		      req_cpu,
		      reqp_cpu,
		      BRdy_cpu,
		      wr_cpu,
		      burst_cpu,
		      
		      // Signals between arbiter and Capim memory-interface
		      data_in_capim,
		      data_out_capim,
		      addr_capim,
		      write_capim,
		      byte_enable_capim,
		      access_capim,
		      intr_capim,
		      inta_capim);
   
   // System pins
   input		clk;
   input		reset;
   
   // Signals between arbiter and mp_mem_interface
   output [31:0] 	MAd_arb;
   output [63:0] 	MDat_in_arb;
   input [63:0] 	MDat_out_arb;
   output		req_arb;
   input		reqp_arb;
   input		BRdy_arb;
   output		wr_arb;
   output 		burst_arb; 		
   
   // Signals between arbiter and CPU
   input [31:0] 	MAd_cpu;
   input [63:0] 	MDat_in_cpu;
   output [63:0] 	MDat_out_cpu;
   input		req_cpu;
   output		reqp_cpu;
   output 		BRdy_cpu;
   input		wr_cpu;
   input 		burst_cpu;
   
   // Signals between arbiter and Capim memory-interface
   input [63:0] 	data_in_capim;
   output [63:0] 	data_out_capim;
   input [21:0] 	addr_capim;
   input  		write_capim;
   input [7:0] 		byte_enable_capim;
   input		access_capim;
   output 		intr_capim;
   input		inta_capim;
   
 	
   reg [3:0] 		next_state;
   wire 		cpu_access;
   reg 			mux_select;
   reg [63:0] 		data_out_capim;
   reg 			intr_capim;
   reg 			burst_capim;   
   
   // these signals are the capim-side input to the muxes
   reg 			reg_req;
   reg 			req_capim;
   reg [21:0] 		MAd_capim;
   reg 			wr_capim;
   reg [63:0] 		MDat_in_capim;
   wire [63:0] 		MDat_out_capim;   
   
   // several muxes to switch between cpu-signals and capim-signals
   assign 		MAd_arb=		(mux_select==`select_cpu) ? MAd_cpu:{10'b0,MAd_capim};
   assign 		MDat_in_arb=	(mux_select==`select_cpu) ? MDat_in_cpu:MDat_in_capim;
   assign 		req_arb=		(mux_select==`select_cpu) ? req_cpu:req_capim;
   assign 		wr_arb=		(mux_select==`select_cpu) ? wr_cpu:wr_capim;
   assign 		burst_arb= (mux_select==`select_cpu) ? burst_cpu:0;   

   // output signals to capim
   assign 		MDat_out_capim=	(mux_select==`select_capim) ? MDat_out_arb:64'b1; /// war bx HHHHH
			
   // output signal to cpu
   assign 		MDat_out_cpu=	(mux_select==`select_cpu) ? MDat_out_arb:64'b1; // was bx HHHH
   assign 		BRdy_cpu=	(mux_select==`select_cpu) ? BRdy_arb:1'b0;
   assign 		reqp_cpu=	(mux_select==`select_cpu) ? reqp_arb:reg_req;


   assign 		cpu_access=(req_cpu | reqp_cpu | BRdy_cpu); 
   
   // state machine
   always @ (posedge clk)
     begin
	if (reset)
	  begin // initial state
	     mux_select <= #100 `select_cpu;
	     reg_req <= #100 `LO;
	     req_capim <= #100 `LO;
	     MAd_capim <= #100 22'bx;
	     MDat_in_capim <= #100 64'bx;
	     intr_capim <= #100 `LO;
	     data_out_capim <= #100 64'bx;	 
	     
	     next_state <= #100 `arb_state_wait4request; 
	  end	
	else
	  case (next_state)
	    `arb_state_wait4request:
	      begin
		 mux_select <= #100 `select_cpu;
		 intr_capim <= #100 `LO;
		 
		 if (access_capim)
		   next_state <= #100 `arb_state_wait4cpu;	     
	      end
	    `arb_state_wait4cpu:
	      begin
		 reg_req <= #100 `LO;
		 
		 if ( (!cpu_access) & write_capim)
		   begin
		      next_state <= #100 `arb_state_write2;
		      mux_select <= #100 `select_capim;
		      
		      req_capim <= #100 `HI;
	     	      MAd_capim <= #100 addr_capim;
	     	      wr_capim <= #100 `HI;
	     	      burst_capim <= #100 `LO;
		   end
		 if ( (!cpu_access) & (!write_capim))
		   begin
		      next_state <= #100 `arb_state_read2;
		      mux_select <= #100 `select_capim;
		      
		      req_capim <= #100 `HI;
	   	      MAd_capim <= #100 addr_capim;
	   	      burst_capim <= #100 `LO;
	   	      wr_capim <= #100 `LO;
		   end
	      end
	    
	    `arb_state_write2:
	      begin
		 MDat_in_capim <= #100 data_in_capim;
		 req_capim <= #100 `LO;


		 if (req_cpu) reg_req<= #100 `HI;	
		 
		 if (BRdy_arb)
  		   next_state <= #100 `arb_state_finish;
	      end
	    `arb_state_finish:
	      begin
		 wr_capim <= #100 `LO;
		 MAd_capim <= #100 22'bx;
		 MDat_in_capim <= #100 64'bx;
		 intr_capim <= #100 `HI;	   
		 
		 if (inta_capim) 
		   begin	
		      next_state <= #100 `arb_state_wait4request;
		      // use xxx_capim wires to simulate a cpu-access which began during capim-access
		      req_capim <= #100 (reg_req | req_cpu); 
		      MAd_capim <= #100 MAd_cpu;
		      wr_capim <= #100 wr_cpu;
		      burst_capim <= #100 burst_cpu;
		   end
	      end
	    `arb_state_read2:
	      begin
		 req_capim <= #100 `LO;
		 
		 if (req_cpu) reg_req<= #100 `HI;	
		 
		 if (BRdy_arb)
		   next_state <= #100 `arb_state_read3;
	      end
	    `arb_state_read3:
	      begin
		 data_out_capim <= #100 MDat_out_capim;
		 
		 if (req_cpu) reg_req<= #100 `HI;	
		 
		 next_state <= #100 `arb_state_finish;
	      end
	  endcase
     end

endmodule





