`define state_wait            2'b00
`define state_write           2'b01
`define state_read            2'b10
`define state_interrupt       2'b11

`define LO		1'b0
`define HI		1'b1

module capim_control(
		     clk,
		     slow_clk,
		     reset,
		     capim1_do,
		     capim2_do,
		     capim3_do,
		     capim4_do,
		     capim1_wr,
		     capim2_wr,
		     capim3_wr,
		     capim4_wr,
		     capim1_inta,
		     capim2_inta,
		     capim3_inta,
		     capim4_inta,
		     capim1_di,
		     capim2_di,
		     capim3_di,
		     capim4_di,
		     capim1_intr,
		     capim2_intr,
		     capim3_intr,
		     capim4_intr,
		     cache_address,
		     cache_dout,
		     cache_mw,
		     cache_mr,
		     cache_mbw,
		     cache_din,
		     cache_busy,
		     cpu_reset
		     );

   input           clk;
   input 	   slow_clk;
   input 	   reset;

   // fast 33MHz signals
   input [31:0]    capim1_do;
   input [31:0]    capim2_do;
   input [31:0]    capim3_do;
   input [31:0]    capim4_do;   
   input 	   capim1_wr;
   input 	   capim2_wr;
   input 	   capim3_wr;
   input 	   capim4_wr;
   input 	   capim1_inta;
   input 	   capim2_inta;
   input 	   capim3_inta;   
   input 	   capim4_inta;
   output [31:0]   capim1_di;
   output [31:0]   capim2_di;
   output [31:0]   capim3_di;
   output [31:0]   capim4_di;
   output 	   capim1_intr;
   output 	   capim2_intr;
   output 	   capim3_intr;
   output 	   capim4_intr;

   // slow 33/4 MHz signals
   output [28:0]   cache_address;   
   output [63:0]   cache_dout;
   output 	   cache_mw;
   output 	   cache_mr;
   output [7:0]    cache_mbw;
   input [63:0]    cache_din;
   input 	   cache_busy;

   // this is 33 MHz clocked (via input_reg4), does not matter
   // if asynchronous to slow clk
   output 	   cpu_reset;

   // die Input-Register zwischen Capims und VAMP
   // clocked by fast clock
   reg [31:0] 	   input_reg1;
   reg [31:0] 	   input_reg2;
   reg [31:0] 	   input_reg3;
   reg [31:0] 	   input_reg4;

   // die Output-Register zwischen VAMP und Capims
   // clocked by slow clock
   reg [31:0] 	   capim1_di;
   reg [31:0] 	   capim2_di;   
   reg [31:0] 	   capim3_di;
   reg [31:0] 	   capim4_di;
   // clocked by fast clock
   reg 		   capim1_intr;
   reg 		   capim2_intr;
   reg 		   capim3_intr;
   reg 		   capim4_intr;
   
   // die Register zum Cache
   // clocked by slow clock
   reg [28:0] 	   cache_address;
   reg [63:0] 	   cache_dout;
   reg 		   cache_mw;
   reg 		   cache_mr;
   reg [7:0] 	   cache_mbw;

   // cpu_reset
   assign 	   cpu_reset=input_reg4[31];   
   
   // das State-Register
   reg [1:0] 	   next_state;

   // delayed Register for sync fast/slow clk
   reg 		   slow_take_it; // clocked fast, indicates that slow clock shall take on new data
   reg 		   slow_got_it;  // clocked slow, indicates that slow clock took new data
   reg 		   generate_intr; // tells the fast clock to generate an interupt on capim3
 		   

   // Die Ausgnge der Capims werden durch Register gefhrt.
   // Diese Register haben als Clock-Enable-Signal capim3_wr
   always @ (posedge clk)
     begin
	// capim1-3 regeln Speicherzugriffe
	if (capim1_wr) input_reg1 <= #100 capim1_do;
	if (capim2_wr) input_reg2 <= #100 capim2_do;
	if (capim3_wr) input_reg3 <= #100 capim3_do;	     
	// capim4 regelt den CPU-Reset
	if (capim4_wr) input_reg4 <= #100 capim4_do;

	// delayed register
	slow_take_it <= #100 (slow_take_it | capim3_wr) & ~slow_got_it;
     end

   // der folgende Automat regelt die Kommunikation mit dem
   // Host-PC ber die Capims, slow-clocked
   always @ (posedge slow_clk)
     begin
	if (reset)
	  begin
	     // reset
	     next_state <= #100 `state_wait;

	     cache_mw <= #100 `LO;
	     cache_mr <= #100 `LO;

	     slow_got_it <= #100 `LO;
	  end
	else
	  begin
	     case (next_state)
	       `state_wait:
		 begin
		    cache_mr <= #100 `LO;
		    cache_mw <= #100 `LO;
		    generate_intr <= #100 `LO;
		    
		    if (slow_take_it)
		      begin
			 slow_got_it <= `HI;
		 
			 cache_address <= #100 {7'b0000000,input_reg3[31:10]};
			 
			 if (input_reg3[9])
			   begin // Schreibzugriff			     
			      cache_dout <= #100 {input_reg1,input_reg2};
			      cache_mw <= #100 `HI;
			      cache_mbw <= #100 input_reg3[8:1];			     
			      
			      next_state <= #100 `state_write;			     
			   end
			 else
			   begin // Lesezugriff
			      cache_mr <= #100 `HI;
			      
			      next_state <= #100 `state_read;			     
			   end
		      end
		 end // case: `state_wait
	       `state_write:
		 begin
		    slow_got_it <= `LO;
		    if (cache_busy==`LO) 
		      begin
			 cache_mw <= #100 `LO;
			 
			 next_state <= #100 `state_interrupt;
		      end
		 end
	       `state_read:
		 begin
		    slow_got_it <= `LO;
		    if (cache_busy==`LO)
		      begin
			 cache_mr <= #100 `LO;

			 capim1_di   <= #100 cache_din[63:32];
			 capim2_di   <= #100 cache_din[31:0];

			 next_state <= #100 `state_interrupt;
		      end
		 end // case: `state_read
	       `state_interrupt:
		 begin
		    generate_intr <= #100 `HI;
		    next_state <= #100 `state_wait;
		 end	      
	     endcase // case(next_state)	     
	  end // else: !if(reset)	
     end // always @ (posedge slowclk)   

   always @ (posedge clk)
     // this requires that inta takes at least one slow_clk cycle
     // after intr; otherwise, two intr could be generated
     begin
	if (generate_intr & ~capim3_intr)
	  begin
	     capim3_intr <= #100 `HI;	  
	  end
	if (capim3_intr & capim3_inta)
	  begin
	     capim3_intr <= #100 `LO;	     
	  end   
     end
   
endmodule // capim_control